@@ -6,7 +6,7 @@
 #
 # The MIT License (MIT)
 #
-# Copyright (c) 2014-2019 Paul Sokolovsky
+# Copyright (c) 2014-2020 Paul Sokolovsky
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -26,6 +26,7 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
 
+from pycopy import const
 import sys
 import ffilib
 import uarray as array
@@ -56,6 +57,8 @@ MULTILINE = M = 2
 DOTALL = S = 4
 VERBOSE = X = 8
 PCRE_ANCHORED = 0x10
+# Ignored
+LOCALE = L = 0
 
 # ASCII is a special value ditinguishable from 0
 ASCII = A = 0x4000
@@ -63,6 +66,11 @@ UNICODE = U = 0x800
 
 PCRE_INFO_CAPTURECOUNT = 2
 
+_UNESCAPE_DICT = const({
+    b"\\": b"\\", b"n": b"\n", b"r": b"\r", b"t": b"\t",
+    b"v": b"\v", b"f": b"\f", b"a": b"\a", b"b": b"\b"
+})
+
 
 class error(Exception):
     pass
@@ -77,8 +85,11 @@ class PCREMatch:
         self.num = num_matches
         self.offsets = offsets
 
-    def substr(self, i):
-        if isinstance(i, str):
+    def _has_group(self, i):
+        return i < len(self.offsets) >> 1
+
+    def substr(self, i, default=None):
+        if isinstance(i, (str, bytes)):
             n = i
             i = pcre_get_stringnumber(self.patobj, n)
             if i < 0:
@@ -86,7 +97,7 @@ class PCREMatch:
         i <<= 1
         beg = self.offsets[i]
         if beg == -1:
-            return None
+            return default
         s = self.s[beg:self.offsets[i + 1]]
         if self.is_str:
             s = s.decode()
@@ -100,8 +111,7 @@ class PCREMatch:
         return tuple(self.substr(i) for i in n)
 
     def groups(self, default=None):
-        assert default is None
-        return tuple(self.group(i + 1) for i in range(self.num - 1))
+        return tuple(self.substr(i + 1, default) for i in range(self.num - 1))
 
     def start(self, n=0):
         return self.offsets[n*2]
@@ -140,21 +150,65 @@ class PCREPattern:
         return self.search(s, pos, endpos, PCRE_ANCHORED)
 
     def _handle_repl_escapes(self, repl, match):
-        def handle_backrefs(bk_match):
-            gr = bk_match.group(1)
-            if gr.startswith("g"):
-                gr = gr[2:-1]
-            gr = int(gr)
-            return match.group(gr)
-        return sub(r"\\(\d+|g<\d+>)", handle_backrefs, repl)
+
+        def handle_escape(esc_match):
+            gr = None
+            e = esc_match.group(1)
+            if e.startswith("0"):
+                # Octal escape
+                return chr(int(e, 8))
+            elif e.startswith("g"):
+                gr = e[2:-1]
+            elif e.isdigit():
+                is_oct = False
+                if e.startswith("0"):
+                    is_oct = True
+                elif len(e) >= 3:
+                    try:
+                        int(e[:3], 8)
+                        is_oct = True
+                    except:
+                        pass
+
+                if is_oct:
+                    rest = e[3:]
+                    e = e[:3]
+                    val = int(e, 8)
+                    if val > 0xff:
+                        raise error(r"octal escape value \%s outside of range 0-0o377" % e.decode(), None, 0)
+                    return chr(val).encode() + rest
+
+                rest = e[2:]
+                e = e[:2]
+                gr = int(e)
+                if match._has_group(gr):
+                    return (match.group(gr) or b"") + rest
+                else:
+                    raise error("invalid group reference")
+            else:
+                r = _UNESCAPE_DICT.get(e)
+                if r is None:
+                    return b"\\" + e
+                return r
+
+            if gr.isdigit():
+                gr = int(gr)
+            return match.group(gr) or b""
+
+        return sub(r"\\(0[0-7]*|\d+|g<.+?>|.)", handle_escape, repl)
 
     def sub(self, repl, s, count=0):
-        res = ""
+        is_str = isinstance(s, str)
+        if is_str:
+            s = s.encode()
+
+        res = b""
         pos = 0
         while True:
             m = self.search(s, pos)
             if not m:
-                return res + s[pos:]
+                res += s[pos:]
+                break
             beg, end = m.span()
             res += s[pos:beg]
             if callable(repl):
@@ -167,16 +221,24 @@ class PCREPattern:
             pos = end
             if beg == end:
                 # Have progress on empty matches
-                res += s[pos]
+                res += s[pos:pos + 1]
                 pos += 1
 
             if count != 0:
                 count -= 1
                 if count == 0:
-                    return res + s[pos:]
+                    res += s[pos:]
+                    break
+
+        if is_str:
+            res = res.decode()
         return res
 
     def split(self, s, maxsplit=0):
+        is_str = isinstance(s, str)
+        if is_str:
+            s = s.encode()
+
         res = []
         while True:
             m = self.search(s)
@@ -185,7 +247,7 @@ class PCREPattern:
                 g = m.group(0)
             if not m or not g:
                 res.append(s)
-                return res
+                break
             beg, end = m.span(0)
             res.append(s[:beg])
             if m.num > 1:
@@ -195,28 +257,52 @@ class PCREPattern:
                 maxsplit -= 1
                 if maxsplit == 0:
                     res.append(s)
-                    return res
+                    break
+
+        if is_str:
+            for x in res:
+                if x is not None:
+                    x.__class__ = str
+
+        return res
 
     def findall(self, s, pos=0, endpos=-1):
         if endpos != -1:
             s = s[:endpos]
+
+        is_str = isinstance(s, str)
+        if is_str:
+            s = s.encode()
+
         res = []
         while True:
             m = self.search(s, pos)
             if not m:
-                return res
+                break
             if m.num == 1:
                 res.append(m.group(0))
             elif m.num == 2:
                 res.append(m.group(1))
             else:
-                res.append(m.groups())
+                # Will be modified inplace, so must be copy of literal
+                x = b""[:]
+                res.append(m.groups(x))
             beg, end = m.span(0)
             pos = end
             if beg == end:
                 # Have progress on empty matches
                 pos += 1
 
+        if is_str:
+            for x in res:
+                if isinstance(x, tuple):
+                    for x1 in x:
+                        x1.__class__ = str
+                else:
+                    x.__class__ = str
+
+        return res
+
     def finditer(self, s, pos=0, endpos=-1):
         if endpos != -1:
             s = s[:endpos]
@@ -233,7 +319,16 @@ class PCREPattern:
                 pos += 1
 
 
+def _check_compiled_flags(flags):
+        if flags:
+            raise ValueError("cannot process flags argument with a compiled pattern")
+
+
 def compile(pattern, flags=0):
+    if isinstance(pattern, PCREPattern):
+        _check_compiled_flags(flags)
+        return pattern
+
     if flags & ASCII:
         flags &= ~ASCII
     else:
@@ -249,13 +344,21 @@ def compile(pattern, flags=0):
 
 
 def search(pattern, string, flags=0):
-    r = compile(pattern, flags)
+    if isinstance(pattern, PCREPattern):
+        _check_compiled_flags(flags)
+        r = pattern
+    else:
+        r = compile(pattern, flags)
     return r.search(string)
 
 
 def match(pattern, string, flags=0):
-    r = compile(pattern, flags | PCRE_ANCHORED)
-    return r.search(string)
+    if isinstance(pattern, PCREPattern):
+        _check_compiled_flags(flags)
+        r = pattern
+    else:
+        r = compile(pattern, flags | PCRE_ANCHORED)
+    return r.match(string)
 
 
 def sub(pattern, repl, s, count=0, flags=0):
@@ -268,7 +371,11 @@ def split(pattern, s, maxsplit=0, flags=0):
     return r.split(s, maxsplit)
 
 def findall(pattern, s, flags=0):
-    r = compile(pattern, flags)
+    if isinstance(pattern, PCREPattern):
+        _check_compiled_flags(flags)
+        r = pattern
+    else:
+        r = compile(pattern, flags)
     return r.findall(s)
 
 def finditer(pattern, s, flags=0):
@@ -284,3 +391,7 @@ def escape(s):
         else:
             res += "\\" + c
     return res
+
+
+def purge():
+    pass
